unit SelComp;

{ -------------------------------------------------
  Author: Elena Lisenenkova,  ll@center.spb.su
  
  TSelectedComponentsExpert - AddIn Expert for Delphi 2,3,4 and C++Builder 1,3. 
  Will be installed as MenuItem in MainMenu - 'Edit|Copy Names'.
  ---------------------------------------------------------------}
{ About Experts - see Lischner, chapter 23 :
  "Secrets of Delphi 2" (ISBN 1-57169-026-3, Ray Lischner, Waite Group) }


interface

uses  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
      Forms, Dialogs, ExptIntf, ToolIntf, VirtIntf, Clipbrd, EditIntf, Menus;

type
  TSelectedComponentsExpert = class(TIExpert)
  private
    ClipMenuItem: TIMenuItemIntf;
    CompList: TStringList;
    procedure InstallAddIn;
  protected  
    procedure OnClick(Sender: TIMenuItemIntf); virtual;
  public
    constructor Create; 
    destructor Destroy; override;
    procedure RunSelCompExpert;
    function EnumProc(const FileName, UnitName, FormName: string): Boolean;
    procedure CopyToClipboard;
    procedure Execute; override;
    function GetName: string; override;
    function GetIDString: string; override;
    function GetStyle: TExpertStyle; override;
    function GetAuthor: string; override;
    function GetComment: string; override;
    function GetMenuText: string; override;
    function GetPage: string; override;
    function GetGlyph: HICON; override;
    function GetState: TExpertState; override;
  end;

function InitSelCompExpert(ToolServices: TIToolServices;
    RegisterProc: TExpertRegisterProc;
    var Terminate: TExpertTerminateProc): Boolean; export;  stdcall;
  
var CF_SELCOMPN: Word;
    
implementation

procedure HandleException;
{ Expert is a DLL, pass the exception to Delphi }
begin
ToolServices.RaiseException(ReleaseException); 
end;

constructor TSelectedComponentsExpert.Create;
begin
inherited Create;  
ClipMenuItem:=nil;
try
  InstallAddIn;
except
  HandleException;
end;
end;

destructor TSelectedComponentsExpert.Destroy;
begin
ClipMenuItem.DestroyMenuItem;
ClipMenuItem.Free;
inherited Destroy;
end;

procedure TSelectedComponentsExpert.InstallAddIn;
var
  MainMenu: TIMainMenuIntf;
  ReferenceMenuItem: TIMenuItemIntf;
  ParentMenu: TIMenuItemIntf;
  Flags: TIMenuFlags;
begin
Flags:=[mfEnabled, mfVisible];  
MainMenu := ToolServices.GetMainMenu;
if MainMenu <> nil then begin
  try
    { I want my MenuItem to be in 'Edit' after 'SelectAll', but
      something wrong takes place in further GetParent for 'EditSelectAllItem'
      in all Delphies (2,3,4) - Delphi either doesn't start at all or
      starts without this MenuItem.
      Delphi bug ???  All other items are OK.
      So I use 'Delete' MenuItem and add 2 to its index.   }
    {ReferenceMenuItem := MainMenu.FindMenuItem('EditSelectAllItem'); {don't use it}
    ReferenceMenuItem := MainMenu.FindMenuItem('EditDeleteItem');  
    if ReferenceMenuItem <> nil then  begin
      try
      ParentMenu := ReferenceMenuItem.GetParent;
      if ParentMenu <> nil then begin
        try
        ClipMenuItem := ParentMenu.InsertItem(ReferenceMenuItem.GetIndex+2,
            'Copy Names', 'CopySelCompItem', 
            'Copy selected component names to Clipboard',0,0,0,Flags,OnClick);    
        finally
        ParentMenu.Free; 
        end;
        end;  { if ParentMenu <> nil  }
      finally
      ReferenceMenuItem.Free; 
      end;
      end; { ReferenceMenuItem <> nil }
  finally
  MainMenu.Free;
  end;  
  end;   {if MainMenu <> nil}
end;

procedure TSelectedComponentsExpert.OnClick(Sender: TIMenuItemIntf);
begin
try
  RunSelCompExpert;
except
  HandleException;
end;
end;

function GetComponentName(Component: TIComponentInterface): string;
begin
  Result := '';
  Component.GetPropValueByName('Name', Result);  
end;

procedure TSelectedComponentsExpert.CopyToClipboard;
const CRLF = #13#10;
var i: Integer;
  L: LongInt;
  Data0: THandle;
  DataPtr0: Pointer;
  Data: THandle;
  DataPtr: Pointer;
  P: PChar;
  S: string;
begin
{ for CF_TEXT : }
L:=0;
for i:=0 to CompList.Count-1  do 
  L:=L+Length(CompList[i]) + 2;
Data := GlobalAlloc(GMEM_MOVEABLE, L+1);
DataPtr := GlobalLock(Data);
P:=PChar(DataPtr);
for i:=0 to CompList.Count-1  do  begin
  S:=CompList[i] + CRLF;
  StrPCopy(P,S);
  P:=P+Length(S);
  end;  

{ for CF_SELCOMPN }
L:=0;
for i:=0 to CompList.Count-1  do 
  L:=L+Length(CompList[i]) + 1;
Data0 := GlobalAlloc(GMEM_MOVEABLE, L+1);
DataPtr0 := GlobalLock(Data0);
P:=PChar(DataPtr0);
for i:=0 to CompList.Count-1  do  begin
  S:=CompList[i] + #0;
  StrPCopy(P,S);
  P:=P+Length(S);
  end;  
StrPCopy(P,#0);  { Double zero at the end }
  
ClipBoard.Open;
Clipboard.SetAsHandle(CF_SELCOMPN, Data0);  
Clipboard.SetAsHandle(CF_Text, Data);  
Clipboard.Close;
GlobalUnlock(Data0);
GlobalUnlock(Data);
{GlobalFree(Data);  in Clipboard unit;  here only if except }
Clipboard.Close;
end;

function TSelectedComponentsExpert.EnumProc(const FileName, UnitName,
  FormName: string): Boolean;
{ is called for every project module in 
  ToolServices.EnumProjectUnits(ProjEnumProc, Self);  }
var ModuleIntf: TIModuleInterface; 
    FormIntf: TIFormInterface;  
    i: Integer;
    CompName: string;
begin
if ToolServices.IsFileOpen(FileName) then 
if FormName <> '' then begin
  ModuleIntf := ToolServices.GetModuleInterface(FileName);
  if ModuleIntf <> nil then begin 
     FormIntf:=ModuleIntf.GetFormInterface;
     if FormIntf <> nil then begin 
       for i:=0 to FormIntf.GetSelCount-1 do begin
         CompName:=GetComponentName(FormIntf.GetSelComponent(i));
         if CompName = FormName then continue; { nothing is selected }
         CompList.Add(FormName+'.'+CompName);
         end;
       FormIntf.Free;
       end; { FormIntf <> nil}
     ModuleIntf.Free;
   end; { ModuleIntf <> nil }
  end;
Result := True;
end;

function ProjEnumProc(Param: Pointer; const FileName, UnitName,
  FormName: string): Boolean; stdcall;
begin
try
  Result := TSelectedComponentsExpert(Param).EnumProc(FileName, UnitName, FormName);
except
  Result := False;
end;
end;

procedure TSelectedComponentsExpert.RunSelCompExpert;
{ all exceptions are handled in caller through HandleException }
begin
CompList:=TStringList.Create;
ToolServices.EnumProjectUnits(ProjEnumProc, Self);
if CompList.Count = 0 then ShowMessage('No components are selected')
else begin
  CopyToClipboard;
  ShowMessage('Selected component names are copied to Clipboard');
  end;
CompList.Free;
CompList:=nil;
end;

function TSelectedComponentsExpert.GetName: string; 
begin
  Result:='Selected Components Expert';
end;

function TSelectedComponentsExpert.GetIDString: string; 
begin
  Result:='LL.SelectedComponentsExpert';
end;

function TSelectedComponentsExpert.GetStyle: TExpertStyle; 
begin
  Result:=esAddIn;
end;

{ not needed for esAddIn  : }

procedure TSelectedComponentsExpert.Execute;
begin
{ is used in Standard Experts }
end;

function TSelectedComponentsExpert.GetState: TExpertState; 
begin
Result := [];
end;

function TSelectedComponentsExpert.GetMenuText: string; 
begin
Result := '';  
end;

function TSelectedComponentsExpert.GetAuthor: string;
begin
Result := 'Elena Lisenenkova';
end;

function TSelectedComponentsExpert.GetComment: string;
begin
Result := '';
end;

function TSelectedComponentsExpert.GetPage: string;
begin
Result := '';
end;

function TSelectedComponentsExpert.GetGlyph: HICON;
begin
Result := 0;
end;  

{ Expert Initialization.  Lischner ch.23 (716-717) }
function InitSelCompExpert(ToolServices: TIToolServices;
    RegisterProc: TExpertRegisterProc;
    var Terminate: TExpertTerminateProc): Boolean; export;  stdcall;
begin
Result := ToolServices <> nil;
if Result then begin
  ExptIntf.ToolServices:=ToolServices;
  Application.Handle:=ToolServices.GetParentHandle;
  Result := RegisterProc(TSelectedComponentsExpert.Create);  { Free in Delphi }
  { no Terminate function}
  end;
end;  

initialization
CF_SELCOMPN := RegisterClipboardFormat('Selected Components');
  
end.
 